#ifndef _HANDLE_DATA_HPP_
#define _HANDLE_DATA_HPP_

#include <chronoptics/tof.h>

#include <functional>
#include <iostream>
#include <mex.hpp>
#include <mexAdapter.hpp>
#include <sstream>

namespace impl {

constexpr size_t user_pointer_size = 100;

using namespace matlab;

inline void error_(std::string message) {
  throw std::runtime_error(std::move(message));
}

inline void check_error_(tof_error_t error) {
  if (!error) return;

  std::string message(tof_error_message(error));
  tof_error_delete(error);

  error_(message);
}

inline std::vector<tof_frame_type> get_frame_types(tof_camera_t camera) {
  tof_error_t error{nullptr};

  // Retrieve set stream list
  size_t capacity = tof_camera_get_set_stream_list(camera, nullptr, 0, &error);
  check_error_(error);

  std::vector<tof_stream_t> streams(capacity, nullptr);
  capacity =
      tof_camera_get_set_stream_list(camera, streams.data(), capacity, &error);
  check_error_(error);

  streams.resize(capacity);

  // Copy over frame types
  std::vector<tof_frame_type> frame_types;
  frame_types.reserve(capacity);
  for (auto stream : streams) {
    auto frame_type = tof_stream_frame_type(stream, &error);
    check_error_(error);
    frame_types.push_back(frame_type);
  }

  // Remove duplicates
  std::sort(frame_types.begin(), frame_types.end());
  auto it = std::unique(frame_types.begin(), frame_types.end());
  frame_types.erase(it, frame_types.end());

  // Clean up allocated data
  for (auto stream : streams) {
    tof_stream_delete(stream);
  }

  return frame_types;
}

inline size_t get_maximum_frame_size(tof_camera_t camera) {
  tof_error_t error{nullptr};

  const auto config = tof_camera_get_camera_config(camera, &error);
  check_error_(error);

  const auto roi = tof_camera_config_get_roi(config, 0, &error);
  check_error_(error);

  const int32_t rows = tof_roi_sensor_rows(roi, &error);
  check_error_(error);

  const int32_t cols = tof_roi_sensor_cols(roi, &error);
  check_error_(error);

  tof_camera_config_delete(config);
  tof_roi_delete(roi);

  return rows * cols;
}

inline size_t get_rgb_frame_size(tof_camera_t camera) {
  tof_error_t error{nullptr};

  const auto config = tof_camera_get_camera_config(camera, &error);
  check_error_(error);

  size_t rows = tof_camera_config_get_rgb_height(config, &error);
  if (error) return 0;

  size_t cols = tof_camera_config_get_rgb_width(config, &error);
  if (error) return 0;

  return rows * cols;
}

class Released {
 public:
  Released(tof_frame_type frame_type, size_t frame_size)
      : frame_type_(frame_type), frame_size_(frame_size) {}

  bool is_released() { return released_; }
  void release() { released_ = true; }

  tof_frame_type frame_type() { return frame_type_; }
  size_t frame_size() { return frame_size_; }

 protected:
 private:
  tof_frame_type frame_type_;
  size_t frame_size_;
  bool released_{false};
};

template <typename T>
class UserPointer : public Released {
 public:
  UserPointer(tof_frame_type frame_type, data::ArrayFactory& factory,
              size_t size)
      : Released(frame_type, size), buffer_(factory.createBuffer<T>(size)) {}

  data::buffer_ptr_t<T> buffer() { return std::move(buffer_); }

 protected:
  data::buffer_ptr_t<T> buffer_;
};

template <typename T>
class CameraHandler : public UserPointer<T> {
 public:
  CameraHandler(tof_camera_t camera, tof_frame_type frame_type,
                data::ArrayFactory& factory, size_t size)
      : UserPointer<T>(frame_type, factory, size), camera_(camera) {
    add_user_pointer();
  }

  tof_user_pointer_destructed_t cb = [](uint8_t*, size_t, void* user_data) {
    auto data = static_cast<CameraHandler<T>*>(user_data);
    if (data->is_released()) {
      delete data;
      return;
    }

    data->add_user_pointer();
  };

 private:
  void add_user_pointer() {
    tof_error_t error{nullptr};
    tof_camera_add_user_pointer(
        camera_, reinterpret_cast<uint8_t*>(this->buffer_.get()),
        this->frame_size() * sizeof(T), cb, this, this->frame_type(), &error);

    if (error) {
      // Failed to add user pointer, clean up
      tof_error_delete(error);
      delete this;
    }
  }

  tof_camera_t camera_;
};

inline void queue_user_pointer(tof_camera_t camera, tof_frame_type frame_type,
                               size_t frame_size, data::ArrayFactory& factory,
                               bool first_queue) {
  switch (frame_type) {
    case TOF_FRAME_TYPE_RAW_COMMON:
    case TOF_FRAME_TYPE_COMMON:
      new CameraHandler<uint16_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_RAW:
      new CameraHandler<int16_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_RAW_AB:
      new CameraHandler<uint16_t>(camera, frame_type, factory,
                                  (first_queue ? frame_size * 2 : frame_size));
      break;
    case TOF_FRAME_TYPE_REAL:
    case TOF_FRAME_TYPE_IMAGINARY:
    case TOF_FRAME_TYPE_PHASE:
    case TOF_FRAME_TYPE_AMPLITUDE:
    case TOF_FRAME_TYPE_X:
    case TOF_FRAME_TYPE_Y:
    case TOF_FRAME_TYPE_Z:
      new CameraHandler<float>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_INTENSITY:
      new CameraHandler<uint8_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_RADIAL:
      new CameraHandler<uint16_t>(camera, frame_type, factory, frame_size);
      break;
    case TOF_FRAME_TYPE_BGR_PROJECTED:
    case TOF_FRAME_TYPE_BGR:
      new CameraHandler<uint8_t>(camera, frame_type, factory,
                                 (first_queue ? frame_size * 3 : frame_size));
      break;
    default: {
      std::ostringstream stream;
      stream << "Frame type " << tof_frame_type_to_string(frame_type, nullptr)
             << " is not supported in matlab";
      error_(stream.str());
      break;
    }
  }
}

inline void tofCameraStart(mex::ArgumentList outputs, mex::ArgumentList inputs,
                           data::ArrayFactory& factory) {
  if (inputs.size() != 2 || outputs.size() != 0) {
    error_("2 inputs are required and 0 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Camera");
  }

  auto camera =
      reinterpret_cast<tof_camera_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  size_t frame_size = get_maximum_frame_size(camera);
  size_t rgb_frame_size = get_rgb_frame_size(camera);

  auto frame_types = get_frame_types(camera);

  tof_camera_set_user_pointer_capacity(camera, user_pointer_size, &error);
  check_error_(error);

  for (auto frame_type : frame_types) {
    auto size =
        (frame_type == TOF_FRAME_TYPE_BGR ? rgb_frame_size : frame_size);

    for (size_t i = 0; i < user_pointer_size; i++) {
      queue_user_pointer(camera, frame_type, size, factory, true);
    }
  }

  // Actually start the camera
  tof_camera_start(camera, &error);
  check_error_(error);
}

inline void tofCameraGetFrames(mex::ArgumentList outputs,
                               mex::ArgumentList inputs,
                               data::ArrayFactory& factory) {
  if (inputs.size() != 2 || outputs.size() != 1) {
    error_("2 inputs are required and 1 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Camera");
  }

  auto camera =
      reinterpret_cast<tof_camera_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  size_t capacity = tof_camera_get_frames(camera, nullptr, 0, &error);
  check_error_(error);
  std::vector<tof_data_t> frames(capacity);
  capacity = tof_camera_get_frames(camera, frames.data(), capacity, &error);
  check_error_(error);
  std::vector<uint64_t> output;
  output.reserve(capacity);
  for (auto item : frames) {
    output.push_back(reinterpret_cast<uint64_t>(item));
  }

  for (auto data : frames) {
    bool user_allocated = tof_data_user_allocated(data, &error);
    check_error_(error);

    if (!user_allocated) continue;

    auto user_pointer = tof_data_user_pointer(data, &error);
    check_error_(error);

    auto base = static_cast<Released*>(user_pointer);
    base->release();

    queue_user_pointer(camera, base->frame_type(), base->frame_size(), factory,
                       false);
  }

  outputs[0] = factory.createArray({1, capacity}, output.data(),
                                   output.data() + capacity);
}

template <typename T>
void create_output(mex::ArgumentList outputs, void* handler_ptr,
                   data::ArrayDimensions& dims, data::ArrayFactory& factory) {
  auto ptr = static_cast<UserPointer<T>*>(handler_ptr);

  auto buffer = ptr->buffer();

  if (buffer.get() == nullptr) {
    throw std::runtime_error("Can only call data once");
  }

  outputs[0] = factory.createArrayFromBuffer(dims, std::move(buffer));
}

inline void set_output(mex::ArgumentList outputs, void* handler_ptr,
                       tof_frame_type frame_type, data::ArrayDimensions& dims,
                       data::ArrayFactory& factory) {
  switch (frame_type) {
    case TOF_FRAME_TYPE_RAW_COMMON:
    case TOF_FRAME_TYPE_COMMON:
      return create_output<uint16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_RAW:
      return create_output<int16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_RAW_AB:
      // dims[0] *= 2;
      return create_output<uint16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_REAL:
    case TOF_FRAME_TYPE_IMAGINARY:
    case TOF_FRAME_TYPE_PHASE:
    case TOF_FRAME_TYPE_AMPLITUDE:
    case TOF_FRAME_TYPE_X:
    case TOF_FRAME_TYPE_Y:
    case TOF_FRAME_TYPE_Z:
      return create_output<float>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_INTENSITY:
      return create_output<uint8_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_RADIAL:
      return create_output<uint16_t>(outputs, handler_ptr, dims, factory);
    case TOF_FRAME_TYPE_BGR_PROJECTED:
    case TOF_FRAME_TYPE_BGR:
      dims.insert(dims.begin(), 3);
      return create_output<uint8_t>(outputs, handler_ptr, dims, factory);
    default: {
      std::ostringstream stream;
      stream << "Frame type " << tof_frame_type_to_string(frame_type, nullptr)
             << " is not supported in matlab";
      error_(stream.str());
      break;
    }
  }
}

inline void tofDataGetData(mex::ArgumentList outputs, mex::ArgumentList inputs,
                           data::ArrayFactory& factory) {
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Data");
  }

  auto data = reinterpret_cast<tof_data_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  bool user_allocated = tof_data_user_allocated(data, &error);
  check_error_(error);

  if (!user_allocated) {
    error_("ToF data is not user allocated");
  }

  auto handler_ptr = tof_data_user_pointer(data, &error);
  check_error_(error);

  size_t rows = tof_data_rows(data, &error);
  check_error_(error);

  size_t cols = tof_data_cols(data, &error);
  check_error_(error);

  auto frame_type = tof_data_frame_type(data, &error);
  check_error_(error);

  data::ArrayDimensions dims{cols, rows};
  set_output(outputs, handler_ptr, frame_type, dims, factory);
}

template <typename T>
inline void copy_data(mex::ArgumentList inputs, void* data_ptr, size_t rows,
                      size_t cols, data::ArrayType type) {
  const size_t size = rows * cols;

  if (inputs[2].getType() != type) {
    error_("Input matrix type does not match frame type");
  }

  if (inputs[2].getNumberOfElements() != size ||
      inputs[2].getDimensions()[0] != cols) {
    error_("Input matrix size does not match");
  }

  const matlab::data::TypedArray<T> arr_t = inputs[2];
  matlab::data::TypedIterator<const T> it(arr_t.begin());
  const T* out_ptr = it.operator->();
  T* in_ptr = reinterpret_cast<T*>(data_ptr);

  std::copy(out_ptr, out_ptr + size, in_ptr);
}

inline void copy_data_select(mex::ArgumentList inputs, void* data_ptr,
                             size_t rows, size_t cols,
                             tof_frame_type frame_type) {
  switch (frame_type) {
    case TOF_FRAME_TYPE_RAW_COMMON:
    case TOF_FRAME_TYPE_COMMON:
    case TOF_FRAME_TYPE_RAW_AB:
    case TOF_FRAME_TYPE_RADIAL:
      return copy_data<uint16_t>(inputs, data_ptr, rows, cols,
                                 data::ArrayType::UINT16);
    case TOF_FRAME_TYPE_RAW:
      return copy_data<int16_t>(inputs, data_ptr, rows, cols,
                                data::ArrayType::INT16);
    case TOF_FRAME_TYPE_REAL:
    case TOF_FRAME_TYPE_IMAGINARY:
    case TOF_FRAME_TYPE_PHASE:
    case TOF_FRAME_TYPE_AMPLITUDE:
    case TOF_FRAME_TYPE_X:
    case TOF_FRAME_TYPE_Y:
    case TOF_FRAME_TYPE_Z:
      return copy_data<float>(inputs, data_ptr, rows, cols,
                              data::ArrayType::COMPLEX_SINGLE);
    case TOF_FRAME_TYPE_INTENSITY:
      return copy_data<uint8_t>(inputs, data_ptr, rows, cols,
                                data::ArrayType::UINT8);
    default: {
      std::ostringstream stream;
      stream << "Frame type " << tof_frame_type_to_string(frame_type, nullptr)
             << " is not supported in setData";
      error_(stream.str());
      break;
    }
  }
}

inline void tofDataSetData(mex::ArgumentList inputs) {
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be Data");
  }

  auto data = reinterpret_cast<tof_data_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  auto ptr = tof_data_data(data, &error);
  check_error_(error);

  size_t rows = tof_data_rows(data, &error);
  check_error_(error);

  size_t cols = tof_data_cols(data, &error);
  check_error_(error);

  auto frame_type = tof_data_frame_type(data, &error);
  check_error_(error);

  copy_data_select(inputs, ptr, rows, cols, frame_type);
}

inline void tofDataData(mex::ArgumentList outputs, mex::ArgumentList inputs,
                        data::ArrayFactory& factory) {
  bool correct_get = inputs.size() == 2 && outputs.size() == 1;
  bool correct_set = inputs.size() == 3 && outputs.size() == 0;

  if (correct_get) {
    tofDataGetData(outputs, inputs, factory);
  } else if (correct_set) {
    tofDataSetData(inputs);
  } else {
    error_("wrong inputs/outputs supplied expecting either 2/1 or 3/0");
  }
}

template <typename T>
class ReaderHandler : public UserPointer<T> {
 public:
  ReaderHandler(tof_frame_type frame_type, data::ArrayFactory& factory,
                size_t size)
      : UserPointer<T>(frame_type, factory, size) {}

  uint8_t* get_pointer() const {
    return reinterpret_cast<uint8_t*>(this->buffer_.get());
  }

  tof_user_pointer_destructed_t cb = [](uint8_t*, size_t, void* user_data) {
    auto data = static_cast<ReaderHandler<T>*>(user_data);
    delete data;
  };
};

template <typename T>
void read_frame(mex::ArgumentList outputs, tof_csf_reader_t reader,
                tof_frame_type frame_type, data::ArrayFactory& factory) {
  tof_error_t error{nullptr};

  auto frame_size = tof_csf_reader_next_frame_size(reader, &error);
  check_error_(error);

  auto handler =
      new ReaderHandler<T>(frame_type, factory, frame_size / sizeof(T));

  auto frame = tof_csf_reader_read_frame_into_pointer(
      reader, handler->get_pointer(), frame_size, handler->cb, handler, &error);
  check_error_(error);

  outputs[0] = factory.createScalar(reinterpret_cast<uint64_t>(frame));
}

inline void read_frame_type(mex::ArgumentList outputs, tof_csf_reader_t reader,
                            tof_frame_type frame_type,
                            data::ArrayFactory& factory) {
  switch (frame_type) {
    case TOF_FRAME_TYPE_RAW_COMMON:
    case TOF_FRAME_TYPE_COMMON:
    case TOF_FRAME_TYPE_RAW_AB:
    case TOF_FRAME_TYPE_RADIAL:
      return read_frame<uint16_t>(outputs, reader, frame_type, factory);
    case TOF_FRAME_TYPE_RAW:
      return read_frame<int16_t>(outputs, reader, frame_type, factory);
    case TOF_FRAME_TYPE_REAL:
    case TOF_FRAME_TYPE_IMAGINARY:
    case TOF_FRAME_TYPE_PHASE:
    case TOF_FRAME_TYPE_AMPLITUDE:
    case TOF_FRAME_TYPE_X:
    case TOF_FRAME_TYPE_Y:
    case TOF_FRAME_TYPE_Z:
      return read_frame<float>(outputs, reader, frame_type, factory);
    case TOF_FRAME_TYPE_INTENSITY:
    case TOF_FRAME_TYPE_BGR_PROJECTED:
    case TOF_FRAME_TYPE_BGR:
      return read_frame<uint8_t>(outputs, reader, frame_type, factory);
    default: {
      std::ostringstream stream;
      stream << "Frame type " << tof_frame_type_to_string(frame_type, nullptr)
             << " is not supported in matlab";
      error_(stream.str());
      break;
    }
  }
}

inline void tofCsfReaderReadFrame(mex::ArgumentList outputs,
                                  mex::ArgumentList inputs,
                                  data::ArrayFactory& factory) {
  if (inputs.size() != 2 || outputs.size() != 1) {
    error_("2 inputs are required and 1 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be CsfReader");
  }

  auto csf_reader =
      reinterpret_cast<tof_csf_reader_t>(static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  auto stream = tof_csf_reader_next_stream(csf_reader, &error);
  check_error_(error);

  auto frame_type = tof_stream_frame_type(stream, &error);
  check_error_(error);

  read_frame_type(outputs, csf_reader, frame_type, factory);
}

class GigeDepthPointer : public UserPointer<uint16_t> {
 public:
  // XXX: Hardcoded maximum sensor size. Might change in the future
  GigeDepthPointer(tof_gige_interface_t interface, data::ArrayFactory& factory)
      : UserPointer<uint16_t>(TOF_FRAME_TYPE_RAW, factory, 640 * 480 * 2),
        interface_(interface) {
    add_user_pointer();
  }

  tof_user_pointer_destructed_t cb = [](uint8_t*, size_t, void* user_data) {
    auto data = static_cast<GigeDepthPointer*>(user_data);
    if (data->is_released()) {
      delete data;
      return;
    }

    data->add_user_pointer();
  };

 private:
  void add_user_pointer() {
    tof_error_t error{nullptr};
    tof_gige_interface_add_depth_user_pointer(
        interface_, reinterpret_cast<uint8_t*>(this->buffer_.get()),
        this->frame_size() * 2, cb, this, &error);

    if (error) {
      // Failed to add user pointer, clean up
      tof_error_delete(error);
      delete this;
    }
  }

  tof_gige_interface_t interface_;
};

inline void tofGigeInterfaceStartDepthStream(mex::ArgumentList outputs,
                                             mex::ArgumentList inputs,
                                             data::ArrayFactory& factory) {
  if (inputs.size() != 4 || outputs.size() != 0) {
    error_("4 inputs are required and 0 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be GigeInterface");
  }
  if (inputs[2].getType() != data::ArrayType::UINT32) {
    error_("Input 3 must be DepthStreamType");
  }
  if (inputs[3].getType() < data::ArrayType::DOUBLE ||
      inputs[3].getType() > data::ArrayType::UINT64) {
    error_("Input 4 must be numerical");
  }

  auto gige_interface = reinterpret_cast<tof_gige_interface_t>(
      static_cast<uint64_t>(inputs[1][0]));

  uint32_t enum_2 = inputs[2][0];
  auto stream = static_cast<tof_depth_stream_type>(enum_2);

  const uint16_t port = inputs[3][0];

  tof_error_t error{nullptr};

  tof_gige_interface_set_user_pointer_capacity(gige_interface,
                                               user_pointer_size, &error);
  check_error_(error);

  for (size_t i = 0; i < user_pointer_size; i++) {
    new GigeDepthPointer(gige_interface, factory);
  }

  tof_gige_interface_start_depth_stream(gige_interface, stream, port, &error);
  check_error_(error);
}

inline void tofGigeInterfaceGetDepthData(mex::ArgumentList outputs,
                                         mex::ArgumentList inputs,
                                         data::ArrayFactory& factory_) {
  if (inputs.size() != 2 || outputs.size() != 1) {
    error_("2 inputs are required and 1 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be GigeInterface");
  }

  auto gige_interface = reinterpret_cast<tof_gige_interface_t>(
      static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  const auto output = tof_gige_interface_get_depth_data(gige_interface, &error);
  check_error_(error);

  auto user_allocated = tof_data_user_allocated(output, &error);
  check_error_(error);

  if (user_allocated) {
    auto user_data = tof_data_user_pointer(output, &error);
    check_error_(error);

    auto pointer = static_cast<GigeDepthPointer*>(user_data);
    pointer->release();

    new GigeDepthPointer(gige_interface, factory_);
  }

  outputs[0] = factory_.createScalar(reinterpret_cast<uint64_t>(output));
}

inline void tofGigeInterfaceGetImageData(mex::ArgumentList outputs,
                                         mex::ArgumentList inputs,
                                         data::ArrayFactory& factory_) {
  if (inputs.size() != 2 || outputs.size() != 1) {
    error_("2 inputs are required and 1 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be GigeInterface");
  }

  auto gige_interface = reinterpret_cast<tof_gige_interface_t>(
      static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  // XXX: Hardcoded maximum sensor size
  const auto size = 1920 * 1080 * 3;
  auto handler = new ReaderHandler<uint8_t>(TOF_FRAME_TYPE_BGR, factory_, size);

  const auto output = tof_gige_interface_get_image_data_into_pointer(
      gige_interface, handler->get_pointer(), size, handler->cb, handler,
      &error);
  check_error_(error);

  outputs[0] = factory_.createScalar(reinterpret_cast<uint64_t>(output));
}

class USBDepthPointer : public UserPointer<uint16_t> {
 public:
  // XXX: Hardcoded maximum sensor size. Might change in the future
  USBDepthPointer(tof_usb_interface_t interface, data::ArrayFactory& factory)
      : UserPointer<uint16_t>(TOF_FRAME_TYPE_RAW, factory, 640 * 480 * 2),
        interface_(interface) {
    add_user_pointer();
  }

  tof_user_pointer_destructed_t cb = [](uint8_t*, size_t, void* user_data) {
    auto data = static_cast<USBDepthPointer*>(user_data);
    if (data->is_released()) {
      delete data;
      return;
    }

    data->add_user_pointer();
  };

 private:
  void add_user_pointer() {
    tof_error_t error{nullptr};
    tof_usb_interface_add_depth_user_pointer(
        interface_, reinterpret_cast<uint8_t*>(this->buffer_.get()),
        this->frame_size() * 2, cb, this, &error);

    if (error) {
      // Failed to add user pointer, clean up
      tof_error_delete(error);
      delete this;
    }
  }

  tof_usb_interface_t interface_;
};

inline void tofUsbInterfaceStartDepthStream(mex::ArgumentList outputs,
                                            mex::ArgumentList inputs,
                                            data::ArrayFactory& factory) {
  if (inputs.size() != 3 || outputs.size() != 0) {
    error_("3 inputs are required and 0 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be UsbInterface");
  }
  if (inputs[2].getType() != data::ArrayType::UINT32) {
    error_("Input 3 must be DepthStreamType");
  }

  auto usb_interface = reinterpret_cast<tof_usb_interface_t>(
      static_cast<uint64_t>(inputs[1][0]));

  uint32_t enum_2 = inputs[2][0];
  auto stream = static_cast<tof_depth_stream_type>(enum_2);

  tof_error_t error{nullptr};

  tof_usb_interface_set_user_pointer_capacity(usb_interface, user_pointer_size,
                                              &error);
  check_error_(error);

  for (size_t i = 0; i < user_pointer_size; i++) {
    new USBDepthPointer(usb_interface, factory);
  }

  tof_usb_interface_start_depth_stream(usb_interface, stream, &error);
  check_error_(error);
}

inline void tofUsbInterfaceGetDepthData(mex::ArgumentList outputs,
                                        mex::ArgumentList inputs,
                                        data::ArrayFactory& factory_) {
  if (inputs.size() != 2 || outputs.size() != 1) {
    error_("2 inputs are required and 1 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be UsbInterface");
  }

  auto usb_interface = reinterpret_cast<tof_usb_interface_t>(
      static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  const auto output = tof_usb_interface_get_depth_data(usb_interface, &error);
  check_error_(error);

  auto user_allocated = tof_data_user_allocated(output, &error);
  check_error_(error);

  if (user_allocated) {
    auto user_data = tof_data_user_pointer(output, &error);
    check_error_(error);

    auto pointer = static_cast<USBDepthPointer*>(user_data);
    pointer->release();

    new USBDepthPointer(usb_interface, factory_);
  }

  outputs[0] = factory_.createScalar(reinterpret_cast<uint64_t>(output));
}

inline void tofUsbInterfaceGetImageData(mex::ArgumentList outputs,
                                        mex::ArgumentList inputs,
                                        data::ArrayFactory& factory_) {
  if (inputs.size() != 2 || outputs.size() != 1) {
    error_("2 inputs are required and 1 outputs are returned");
  }
  if (inputs[1].getType() != data::ArrayType::UINT64) {
    error_("Input 2 must be UsbInterface");
  }

  auto usb_interface = reinterpret_cast<tof_usb_interface_t>(
      static_cast<uint64_t>(inputs[1][0]));

  tof_error_t error{nullptr};

  // XXX: Hardcoded maximum sensor size
  const auto size = 1920 * 1080 * 3;
  auto handler = new ReaderHandler<uint8_t>(TOF_FRAME_TYPE_BGR, factory_, size);

  const auto output = tof_usb_interface_get_image_data_into_pointer(
      usb_interface, handler->get_pointer(), size, handler->cb, handler,
      &error);
  check_error_(error);

  outputs[0] = factory_.createScalar(reinterpret_cast<uint64_t>(output));
}

}  // namespace impl

#endif
