#include <ros/console.h>
#include <ros/init.h>
#include <ros/node_handle.h>
#include <sensor_msgs/CameraInfo.h>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/PointCloud2.h>
#include <sensor_msgs/distortion_models.h>
#include <sensor_msgs/image_encodings.h>

#include <boost/optional.hpp>
#include <chronoptics/tof.hpp>
#include <thread>

using namespace chronoptics;

static std::vector<sensor_msgs::PointField> create_point_fields(
    tof::FrameType frame_type) {
  size_t nr_fields = [frame_type]() {
    switch (frame_type) {
      case tof::FrameType::XYZ:
        return 3;
      case tof::FrameType::XYZ_AMP:
      case tof::FrameType::XYZ_BGR:
        return 4;
      case tof::FrameType::XYZ_BGR_I:
        return 5;
      default:
        throw std::runtime_error("Type not supported for point cloud");
    }
  }();

  std::vector<sensor_msgs::PointField> fields(nr_fields);
  fields[0].name = "x";
  fields[0].offset = 0;
  fields[0].datatype = sensor_msgs::PointField::FLOAT32;
  fields[0].count = 1;

  fields[1].name = "y";
  fields[1].offset = 4;
  fields[1].datatype = sensor_msgs::PointField::FLOAT32;
  fields[1].count = 1;

  fields[2].name = "z";
  fields[2].offset = 8;
  fields[2].datatype = sensor_msgs::PointField::FLOAT32;
  fields[2].count = 1;

  if (frame_type == tof::FrameType::XYZ_AMP) {
    fields[3].name = "intensity";
    fields[3].offset = 12;
    fields[3].datatype = sensor_msgs::PointField::FLOAT32;
    fields[3].count = 1;

  } else if (frame_type != tof::FrameType::XYZ) {
    fields[3].name = "rgb";
    fields[3].offset = 12;
    fields[3].datatype = sensor_msgs::PointField::UINT32;
    fields[3].count = 1;

    if (frame_type == tof::FrameType::XYZ_BGR_I) {
      fields[4].name = "intensity";
      fields[4].offset = 15;
      fields[4].datatype = sensor_msgs::PointField::UINT8;
      fields[4].count = 1;
    }
  }

  return fields;
}

struct PubStreaming {
  ros::Publisher pub;
  bool streaming;
};

struct FramePublishers {
  ros::Publisher camera_info;
  PubStreaming radial;
  PubStreaming intensity;
  PubStreaming z;

  PubStreaming xyz;
  PubStreaming xyzi;

  boost::optional<ros::Publisher> camera_info_rgb;
  boost::optional<PubStreaming> bgr;
  boost::optional<PubStreaming> bgr_projected;

  boost::optional<PubStreaming> xyzbgr;
  boost::optional<PubStreaming> xyzbgri;
};

static bool stream_changed(PubStreaming& publisher) {
  bool now_streaming = publisher.pub.getNumSubscribers() > 0;

  if (now_streaming != publisher.streaming) {
    ROS_INFO("%s %sstreaming", publisher.pub.getTopic().c_str(),
             (now_streaming ? "is now " : "stopped "));
    publisher.streaming = now_streaming;
    return true;
  }

  return false;
}

static void update_streams(tof::Camera& cam, FramePublishers& publishers,
                           std::vector<tof::FrameType>& frame_types) {
  bool update_stream_list{false};

  auto update_stream = [&](PubStreaming& publisher, tof::FrameType frame_type) {
    if (!stream_changed(publisher)) return;

    update_stream_list = true;

    if (publisher.streaming) {
      // Add to frame_types
      frame_types.push_back(frame_type);
    } else {
      // Remove from frame_types
      auto it = std::find(frame_types.begin(), frame_types.end(), frame_type);
      if (it == frame_types.end()) {
        // We should never get here
        assert(false);
        return;
      }

      frame_types.erase(it);
    }
  };

  update_stream(publishers.radial, tof::FrameType::RADIAL);
  update_stream(publishers.intensity, tof::FrameType::INTENSITY);
  update_stream(publishers.z, tof::FrameType::Z);

  update_stream(publishers.xyz, tof::FrameType::XYZ);
  update_stream(publishers.xyzi, tof::FrameType::XYZ_AMP);

  if (publishers.bgr) {
    update_stream(*publishers.bgr, tof::FrameType::BGR);
  }

  if (publishers.bgr_projected) {
    update_stream(*publishers.bgr_projected, tof::FrameType::BGR_PROJECTED);
  }

  if (publishers.xyzbgr) {
    update_stream(*publishers.xyzbgr, tof::FrameType::XYZ_BGR);
  }

  if (publishers.xyzbgri) {
    update_stream(*publishers.xyzbgri, tof::FrameType::XYZ_BGR_I);
  }

  if (!update_stream_list) return;

  tof::select_streams(cam, frame_types);

  // Start/stop camera when required
  if (frame_types.empty()) {
    ROS_INFO("Stopping camera");
    cam.stop();
  } else if (!cam.is_streaming()) {
    ROS_INFO("Starting camera");
    cam.start();
  }
}

static bool has_camera_rgb(tof::Camera& cam) {
  const auto streams = cam.get_stream_list();

  auto it = std::find_if(streams.begin(), streams.end(),
                         [](const tof::Stream& stream) {
                           return stream.frame_type() == tof::FrameType::BGR;
                         });

  return it != streams.end();
}

static FramePublishers create_publishers(ros::NodeHandle& n,
                                         const bool has_rgb) {
  FramePublishers publishers{
      n.advertise<sensor_msgs::CameraInfo>("depth/camera_info", 1),
      {n.advertise<sensor_msgs::Image>("depth/image", 1), false},
      {n.advertise<sensor_msgs::Image>("depth/intensity", 1), false},
      {n.advertise<sensor_msgs::Image>("depth/image_rect", 1), false},
      {n.advertise<sensor_msgs::PointCloud2>("point_cloud/xyz", 1), false},
      {n.advertise<sensor_msgs::PointCloud2>("point_cloud/xyzi", 1), false}};

  if (has_rgb) {
    publishers.camera_info_rgb =
        n.advertise<sensor_msgs::CameraInfo>("rgb/camera_info", 1);
    publishers.bgr = PubStreaming{
        n.advertise<sensor_msgs::Image>("rgb/image_color", 1), false};
    publishers.bgr_projected = PubStreaming{
        n.advertise<sensor_msgs::Image>("depth/image_color_rect", 1), false};

    publishers.xyzbgr = PubStreaming{
        n.advertise<sensor_msgs::PointCloud2>("point_cloud/xyzrgb", 1), false};
    publishers.xyzbgri = PubStreaming{
        n.advertise<sensor_msgs::PointCloud2>("point_cloud/xyzrgbi", 1), false};
  }

  return publishers;
}

struct CallbackImage {
  CallbackImage(tof::Camera& cam_, tof::FrameType frame_type_, uint32_t width,
                uint32_t height, const std::string& encoding)
      : cam(cam_), frame_type(frame_type_), image(new sensor_msgs::Image{}) {
    uint32_t byte_depth = sensor_msgs::image_encodings::bitDepth(encoding) / 8;
    uint32_t nr_columns = sensor_msgs::image_encodings::numChannels(encoding);

    image->width = width;
    image->height = height;
    image->step = width * byte_depth * nr_columns;
    image->is_bigendian = false;
    image->encoding = encoding;

    image->data.resize(height * image->step);
    add_user_pointer();
  }

  tof::Camera& cam;
  tof::FrameType frame_type;

  // Use shared pointer here so that if node is used in nodelet we don't have to
  // serialize/deserialize data
  sensor_msgs::Image::Ptr image;

  tof_user_pointer_destructed_t cb = [](uint8_t*, size_t, void* user_data) {
    auto data = static_cast<CallbackImage*>(user_data);
    // Add data back in, this might cause issues when this node is used in a
    // nodelet. In that case you might want to make the queue_size larger
    try {
      data->add_user_pointer();
    } catch (std::exception& e) {
      ROS_ERROR("%s", e.what());
      delete data;
    }
  };

  void add_user_pointer() {
    cam.add_user_pointer(image->data.data(), image->data.size(), cb, this,
                         frame_type);
  }
};

struct CallbackPointCloud {
  CallbackPointCloud(tof::Camera& cam_, tof::FrameType frame_type_,
                     uint32_t width, uint32_t height)
      : cam(cam_),
        frame_type(frame_type_),
        point_cloud(new sensor_msgs::PointCloud2{}) {
    point_cloud->fields = create_point_fields(frame_type);
    point_cloud->width = width;
    point_cloud->height = height;
    point_cloud->is_bigendian = false;
    point_cloud->is_dense = true;
    point_cloud->point_step = 16;
    point_cloud->row_step = width * 16;

    point_cloud->data.resize(width * height * 16);
    add_user_pointer();
  }

  tof::Camera& cam;
  tof::FrameType frame_type;

  // Use shared pointer here so that if node is used in nodelet we don't have to
  // serialize/deserialize data
  sensor_msgs::PointCloud2::Ptr point_cloud;

  tof_user_pointer_destructed_t cb = [](uint8_t*, size_t, void* user_data) {
    auto data = static_cast<CallbackPointCloud*>(user_data);
    // Add data back in, this might cause issues when this node is used in a
    // nodelet. In that case you might want to make the queue_size larger
    try {
      data->add_user_pointer();
    } catch (std::exception& e) {
      ROS_ERROR("%s", e.what());
      delete data;
    }
  };

  void add_user_pointer() {
    cam.add_user_pointer(point_cloud->data.data(), point_cloud->data.size(), cb,
                         this, frame_type);
  }
};

void queue_user_pointers(tof::Camera& cam, const size_t queue_size,
                         const bool has_rgb) {
  cam.set_user_pointer_capacity(queue_size);
  auto roi = cam.get_camera_config().get_roi(0);

  uint32_t width = roi.get_img_cols();
  uint32_t height = roi.get_img_rows();

  auto queue_image = [&](tof::FrameType frame_type,
                         const std::string& encoding) {
    for (size_t i = 0; i < queue_size; i++) {
      new CallbackImage(cam, frame_type, width, height, encoding);
    }
  };

  queue_image(tof::FrameType::RADIAL, sensor_msgs::image_encodings::TYPE_16UC1);
  queue_image(tof::FrameType::INTENSITY,
              sensor_msgs::image_encodings::TYPE_8UC1);
  queue_image(tof::FrameType::Z, sensor_msgs::image_encodings::TYPE_32FC1);

  if (has_rgb) {
    auto cam_config = cam.get_camera_config();
    uint32_t rgb_width = cam_config.get_rgb_width();
    uint32_t rgb_height = cam_config.get_rgb_height();

    for (size_t i = 0; i < queue_size; i++) {
      new CallbackImage(cam, tof::FrameType::BGR, rgb_width, rgb_height,
                        sensor_msgs::image_encodings::BGR8);
    }

    queue_image(tof::FrameType::BGR_PROJECTED,
                sensor_msgs::image_encodings::BGR8);
  }

  auto queue_point_cloud = [&](tof::FrameType frame_type) {
    for (size_t i = 0; i < queue_size; i++) {
      new CallbackPointCloud(cam, frame_type, width, height);
    }
  };

  queue_point_cloud(tof::FrameType::XYZ);
  queue_point_cloud(tof::FrameType::XYZ_AMP);

  if (has_rgb) {
    queue_point_cloud(tof::FrameType::XYZ_BGR);
    queue_point_cloud(tof::FrameType::XYZ_BGR_I);
  }
}

void update_header(tof::Data& frame, std_msgs::Header& hdr) {
  hdr.stamp = ros::Time::now();
  hdr.seq = frame.frame_count();
  hdr.frame_id = std::to_string(frame.frame_id());
}

void publish_image(ros::Publisher& publisher, tof::Data& frame) {
  if (!frame.user_allocated()) {
    ROS_ERROR("Frame not allocated dropping data");
    // TODO: Copy data
  } else {
    auto cb = static_cast<CallbackImage*>(frame.user_pointer());
    update_header(frame, cb->image->header);
    publisher.publish(cb->image);
  }
}

void publish_pointcloud(ros::Publisher& publisher, tof::Data& frame) {
  if (!frame.user_allocated()) {
    ROS_ERROR("Frame not allocated dropping data");
    // TODO: Copy data
  } else {
    auto cb = static_cast<CallbackPointCloud*>(frame.user_pointer());
    update_header(frame, cb->point_cloud->header);

    // Turn xyz data into meters
    float* data = reinterpret_cast<float*>(cb->point_cloud->data.data());
    const size_t size = frame.rows() * frame.cols();
    for (size_t i = 0; i < size; i++) {
      data[i * 4] /= 1000.f;      // x
      data[i * 4 + 1] /= 1000.f;  // y
      data[i * 4 + 2] /= 1000.f;  // z
      // Don't divide the fourth piece of data since that won't be x, y or z
    }

    publisher.publish(cb->point_cloud);
  }
}

void publish_frames(FramePublishers& publishers,
                    std::vector<tof::Data>& frames) {
  for (auto& frame : frames) {
    switch (frame.frame_type()) {
      case tof::FrameType::RADIAL:
        publish_image(publishers.radial.pub, frame);
        break;
      case tof::FrameType::INTENSITY:
        publish_image(publishers.intensity.pub, frame);
        break;
      case tof::FrameType::Z:
        publish_image(publishers.z.pub, frame);
        break;
      case tof::FrameType::BGR:
        publish_image(publishers.bgr->pub, frame);
        break;
      case tof::FrameType::BGR_PROJECTED:
        publish_image(publishers.bgr_projected->pub, frame);
        break;

      case tof::FrameType::XYZ:
        publish_pointcloud(publishers.xyz.pub, frame);
        break;
      case tof::FrameType::XYZ_AMP:
        publish_pointcloud(publishers.xyzi.pub, frame);
        break;
      case tof::FrameType::XYZ_BGR:
        publish_pointcloud(publishers.xyzbgr->pub, frame);
        break;
      case tof::FrameType::XYZ_BGR_I:
        publish_pointcloud(publishers.xyzbgri->pub, frame);
        break;

      default:
        ROS_ERROR("Unexpected tof frame type");
        break;
    }
  }
}

static sensor_msgs::CameraInfo create_depth_camera_info(tof::Camera& cam) {
  sensor_msgs::CameraInfo cam_info;

  auto camera_config = cam.get_camera_config();
  auto calibration = cam.get_calibration();
  auto roi = camera_config.get_roi(0);

  cam_info.width = roi.sensor_cols();
  cam_info.height = roi.sensor_rows();

  cam_info.distortion_model = sensor_msgs::distortion_models::PLUMB_BOB;

  auto d = calibration.get_depth_distortion_coefficients();
  cam_info.D.insert(cam_info.D.end(), d.begin(), d.end());

  auto k = calibration.get_depth_camera_matrix();
  std::copy(k.begin(), k.end(), cam_info.K.begin());

  cam_info.binning_x = camera_config.get_binning(0) + 1;
  cam_info.binning_y = cam_info.binning_x;

  cam_info.roi.width = roi.get_img_cols();
  cam_info.roi.height = roi.get_img_rows();
  cam_info.roi.x_offset = roi.get_col_offset();
  cam_info.roi.y_offset = roi.get_row_offset();

  return cam_info;
}

static sensor_msgs::CameraInfo create_rgb_camera_info(tof::Camera& cam) {
  sensor_msgs::CameraInfo cam_info;

  auto camera_config = cam.get_camera_config();
  auto calibration = cam.get_calibration();

  cam_info.width = camera_config.get_rgb_width();
  cam_info.height = camera_config.get_rgb_height();

  cam_info.distortion_model = sensor_msgs::distortion_models::PLUMB_BOB;

  auto d = calibration.get_rgb_distortion_coefficients();
  cam_info.D.insert(cam_info.D.end(), d.begin(), d.end());

  auto k = calibration.get_rgb_camera_matrix();
  std::copy(k.begin(), k.end(), cam_info.K.begin());

  return cam_info;
}

void log_tof_cb(int16_t level, const char* log_name, const char* message) {
  if (level == 2) {
    ROS_INFO("%s: %s", log_name, message);
  } else if (level == 3) {
    ROS_WARN("%s: %s", log_name, message);
  } else if (level >= 4) {
    ROS_ERROR("%s: %s", log_name, message);
  }
}

int main(int argc, char** argv) {
  ros::init(argc, argv, "kea_camera");

  std::thread spin{[]() { ros::spin(); }};

  ros::NodeHandle n("~");

  // Serial can be empty
  auto serial = n.param<std::string>("serial", "");

  // Config file can be empty
  auto config_file = n.param<std::string>("config_file", "");

  auto log = n.param<bool>("log", true);

  size_t queue_size = n.param<int>("queue_size", 10);

  bool on_camera_processing = n.param<bool>("on_camera_processing", false);

  // log_fun needs to stay in existence
  std::function<void(int16_t, const char*, const char*)> log_fun = log_tof_cb;
  if (log) tof::log_callback_separated(log_fun);

  tof::ProcessingConfig config(config_file);
  // Set radial scale to 1.0 such that radial distance is in mm
  config.set_radial_scale(1.0);

  // Can be replaced with CsfCamera or EmbeddedKeaCamera
  tof::KeaCamera cam(config, serial);

  if (on_camera_processing && !cam.on_camera_processing_capable()) {
    ROS_ERROR("Camera is not able to do on camera processing");
  } else {
    cam.set_on_camera_processing(on_camera_processing);
  }

  ROS_INFO("Connected to camera %s", cam.get_serial());

  const bool has_rgb = has_camera_rgb(cam);
  auto publishers = create_publishers(n, has_rgb);

  auto depth_cam_info = create_depth_camera_info(cam);
  sensor_msgs::CameraInfo rgb_cam_info;
  if (has_rgb) rgb_cam_info = create_rgb_camera_info(cam);

  queue_user_pointers(cam, queue_size, has_rgb);

  std::vector<tof::FrameType> current_frame_types;
  ros::Rate rate(100.0);
  while (ros::ok() && cam.is_connected()) {
    update_streams(cam, publishers, current_frame_types);

    if (cam.is_streaming()) {
      auto frames = cam.get_frames();

      publishers.camera_info.publish(depth_cam_info);
      if (has_rgb) publishers.camera_info_rgb->publish(rgb_cam_info);

      publish_frames(publishers, frames);
    }

    rate.sleep();
  }

  ros::shutdown();
  spin.join();
}